"Python3 evolutionary regression-rule finder, as part of RUNSTER."

##  ROOT :
##
##  Regression-Oriented Optimization Tester.
##  Copyright (c), 2016, R.S. Forsyth.
##  ancestral progenitor : Sept, 1980.
##  first version : 15/08/2015
##  last revision : 12/09/2016


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  copy
import  math
import  os, sys
import  random, time

##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  pybyebye
from glib import  request
from gridlib import  cols2use, matread
from gridlib import  matdump, vecline
from gridlib import  var2col
from gridlib import  bastatz, midvalue
from gridlib import  corsums, corrval
from gridlib import  ranx
from gridlib import  madm
from gridlib import  fivenum
from gridlib import  vecdist
from gridlib import  copygrid, distest
from gridlib import  matshow, precrec
from gridlib import  sckappa2
from gridlib import  hedgesg
from evosoft import  evoprep
from evosoft import  initpop
from evosoft import  evotries
import  rulelib
import  treelib
import  tidylib


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ##  ("atomize",1,0,1),
    ("ntrials",131072,256,1048576),  ##  total number of evolutionary trials
    ("randseed",0,0,999999999),  ##  random seed, <2 means haphazard seed
    ("jobname","runster"),  ##  name to link related files
    ("casefold",0,0,1),  ##  whether to fold to lower case
    ("traindat",''),  ##  file for input of data
    ("longtime",3600,600,86400),  ##  secs indicating excessive elapsed time
    ##  ("cols2use",''),  ##  variables to be used [set internally]
    ("brevity",1,0,1),  ##  brevity bias selector
    ("targval",''),  ##  target epression or variable
    ("targcol",-1),  ##  target column number [2b reset]
    ("idfield",''),  ##  varname for o/p label
    ("skipvars",''),  ##  variables to be excluded
    ("popsize",233,8,2048),  ##  population size
    ("m1",0.5,0,1),  ##  mutation rate 1
    ("m2",0.25,0,1),  ##  mutation rate 2
    ("rulemode","standard"),  ##  rule interpretation mode, standard or tabular
    ("trigfunx",0,0,1),  ##  whether to allow trig functions Cosi,Sine
    ("listfile",''),  ##  main output file
    ("dumpfile",''),  ##  output file for subsampling results
    ("rulefile",''),  ##  file for o/p of rules
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"root.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "Regression-Oriented Optimization Tester, v2.3"
##  Gold = 5.0**0.5*0.5+0.5
Mincases = 4
Dattail = "_dat1"
Dumptail = "_dump"
Listtail = "_list"
Ruletail = "_rule"
Nicereps = 144  ##  ideal number of subsampling cases
Maxcats = 64
Dustbin = "??!!"
Longname = 26  ##  length after which category names will be truncated
Longlab = 20  ##  length after which label fields will be truncated
Outpars = ["dateline","progname","id","traindat","testdat","targval"]
Goodsign = '+'
Poorsign = '-'


##  Main program :
##
def main():
    "Main line."
    
    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time() ; clockon = time.clock()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        pybyebye()
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()

    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if not paradat.outpath.endswith(os.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
        if not os.path.isdir(paradat.outpath):  ##  try another outpath
            paradat.outpath = os.getcwd() + os.sep + "op" + os.sep
            if not os.path.isdir(paradat.outpath):
                os.makedirs(paradat.outpath)  ##  create o/p folder
            paradat.outfile = paradat.outpath + os.path.basename(paradat.outfile)

    if paradat.randseed > 1:  ##  initialize early
        random.seed(paradat.randseed)  ##  repeatable start
    else:
        random.seed(None)  ##  haphazard start
        
    ##  locate & read input file :
    ##  print(paradat.traindat)
    tf = ''
    if paradat.traindat and os.path.isfile(paradat.traindat):
        print("data to be read from",paradat.traindat)
        r = getfile(paradat,paradat.traindat)
        tf = paradat.traindat
    else:
        tf = paradat.outpath + paradat.jobname + Dattail + ".dat"
        print("data to be sought on",tf)
        if os.path.isfile(tf):
            r = getfile(paradat,tf)
        else:
            print("No data file found!")
            pybyebye() ; return  -2
    if r < 0:
        if not tf:
            tf = paradat.traindat
        print("Problem reading data file "+tf)
        pybeybye() ; return  -2
    ##  vars 2b skipped will have already affected paradat.goodcols.
    paradat.traindat = os.path.abspath(tf)  ##  record name if successful.
    print(paradat.traindat,"read in.")
    print(paradat.datamat[0]) ; print("..")
    print(paradat.datamat[-1])

    ##  establish target values :
    if paradat.targval:
        targets(paradat)
        ##  should make .targvals & update .cols2use
    else:
        print("No target expression (targval) given!!")
        pybyebye()
        return  -3
    ##  check that they're numeric :
    nums = [isinstance(tv,(float,int)) for tv in paradat.targvals]
    nums = sum(nums)
    ##  print(nums)
    if nums < paradat.rows:  ##  nonnumeric
        print("target values must be numeric!!")
        pybyebye() ; return  -4
    tvalues = sorted(set(paradat.targvals))

    ##  prepare o/p files :
    if paradat.listfile:
        listname = paradat.listfile
    else:
        listname = paradat.outpath + paradat.jobname + Listtail + ".txt"
    paradat.listfile = open(listname,'w',encoding="utf8")
    if paradat.rulefile:
        rf = paradat.rulefile
    else:
        rf = paradat.outpath + paradat.jobname + Ruletail + ".txt"
    paradat.rulefile = open(rf,'w',encoding="utf-8")
    if paradat.dumpfile:
        df = paradat.dumpfile
    else:
        df = paradat.outpath + paradat.jobname + Dumptail + ".txt"
    paradat.dumpfile = open(df,'w',encoding="utf-8")
    
    ##  identify categories :
    catlist = []
    catfrex = {}
    for r in range(paradat.rows):
        ##  tv = paradat.datamat[r][paradat.targcol]
        tv = paradat.targvals[r]
        if tv in catfrex:
            catfrex[tv] += 1
        else:
            catfrex[tv] = 1  ## first encounter
        if not (tv in catlist):
            catlist.append(tv)
    paradat.catlist = sorted(catlist)
    paradat.cats = len(paradat.catlist)
    rootrows = paradat.rows ** 0.5
    if paradat.cats < 2:
        print("Insufficient target variation to continue!!")
        pyebye() ; return  -5
    elif paradat.cats < rootrows:
        print("Only",paradat.cats,"distinct target values, as follows :")
        print("Target values :",paradat.catlist)
        print("Are you sure you want to do regression rather than classification?")
    else:
        print(paradat.cats,"target values.")
    paradat.catfrex = catfrex
    ##  paradat.guessing = 0.5
    
    ##  data prescans to find some plausible constants :
    ##  establish cutpoints (& target stats) :
    pickcuts(paradat)  ##  also makes paradat.targstat
    print(paradat.cutvals,file=paradat.dumpfile)
    showobj(paradat.targstat)
    showobj(paradat.targstat,paradat.dumpfile)
    paradat.targmean = paradat.targstat.mean
    ##  tvalues = sorted(set(paradat.targvals))
    desired = round(math.sqrt(paradat.rows)+math.log1p(paradat.rows))
    if desired < len(tvalues):
        paradat.targcuts = sorted(random.sample(tvalues,desired))
    else:
        paradat.targcuts = tvalues
    paradat.targnorm = midvalue(tvalues)
    ##  targnorm still in use. [2b replaced?]
    x = sum(paradat.targcuts)
    nc = len(paradat.targcuts)
    print("targcuts",paradat.targcuts,file=paradat.dumpfile)
    ##  print(x,nc,x/nc,paradat.targnorm,paradat.targmean)
    ##  pseudo-priors :
    paradat.priorvec = [paradat.rows,paradat.targmean,paradat.targstat.std1]
    ##  decide on small difference for 'success' indicator :
    paradat.smalldif = paradat.targstat.madm
    ##  ps = paradat.targstat.std1 * 0.3148
    ##  print(paradat.smalldif,ps)
    paradat.priorvec.append(paradat.smalldif)
    ##  print(paradat.smalldif)
    ##  compute Tukey's five numbers as targcuts :
    n5 = fivenum(tvalues)
    if len(n5) < 5:
        n5.append(paradat.targmean)
    n5.sort()  ##  tidier in order
    paradat.targcuts = n5  ##  2b passed over in rule
    print("targcuts :",paradat.targcuts,file=paradat.dumpfile)

    ##  establish constants for each string variable :
    pickstrs(paradat)
    ##  .strcols & .strdict updated.
    ##  ensure appropriate columns in play :
    if hasattr(paradat,"skipvars"):
        paradat.skipvars = paradat.skipvars.split(',')
        paradat.skipvars = [sv.strip() for sv in paradat.skipvars]
        ##  whitespace cd be a problem.
    else:
        paradat.skipvars = []
    ##  convert .cols2use into a list :
    paradat.cols2use = sorted(paradat.cols2use)
    ##  belt & braces :
    ##  print(paradat.skipvars) ; print(paradat.cols2use)
    for c in range(paradat.cols):
        cn = paradat.taglist[c]
        if cn in paradat.skipvars and c in paradat.cols2use:
            paradat.cols2use.remove(c)
    ##  skipvars ignored strcols on input.
    for c in range(paradat.cols):
        cn = paradat.taglist[c]
        if c in paradat.cols2use:
            cx = '+'
        else:
            cx = '-'
        print(c,cx,cn,file=paradat.dumpfile)
    ##  strcheck(paradat)
    ##  treetest(paradat)

    treelib.trigset(paradat.trigfunx)
    ##  decides whether to include trig functions.
    ##  determine suitable stab size :
    stabprep(paradat)
    ##  showobj(paradat)

    ##  prepare main loop :
    ##  pass info over to evosoft :
    evostuff = Sack("Evostuff")
    evostuff.makerule = rulelib.makerule
    evostuff.copyrule = rulelib.copyrule
    ##  pick suitable fitness function :
    if paradat.rulemode == "tabular":
        evostuff.fitness = rulelib.qualitab
    else:  ##  default is standard
        evostuff.fitness = rulelib.regscore  ##  .quickfit?
    evostuff.mate = rulelib.mate
    evostuff.clobber = rulelib.zonk

    ##  compute cycles & subsample size per cycle :
    nc,testsize = cyclist(paradat.rows,Nicereps)
    paradat.mainreps = nc
    print(nc,testsize)
    paradat.subtries = round(paradat.ntrials / nc)  ##  test cases in each loop
    print("evotrials per cycle =",paradat.subtries)
    print(nc*paradat.subtries)
    leftover = paradat.ntrials - nc*paradat.subtries
    print("leftover =",leftover)
    testnum = nc * testsize  ##  number of subsampling cases
    evostuff.ntrials = paradat.subtries
    evostuff.m1 = paradat.m1 ; evostuff.m2 = paradat.m2
    evostuff.popsize = paradat.popsize
    ##  showobj(evostuff)
    evoprep(evostuff)  ##  transfer to evosoft

    ##  prepare subsampling loop :
    print("number of evolutionary loops =",paradat.mainreps)
    print("subsampling trials to be made =",testnum)
    ncycles = 0
    rownums = list(range(paradat.rows))  ##  all indices
    notused = set(rownums)  ##  indices of never-used cases
    results = []
    indexes = randsubs(paradat.rows,testsize,paradat.mainreps)
    ##  shuffled rownums, followed if necessary by random subsets.
    ##  print(len(indexes)) ; print(indexes)
    cycling = "cycle"  ##  singular
    used = set()
    toprules = []  ##  list of best in each cycle

    ##  main train/test loop :
    while len(results) < testnum:
        ##  select training & test subsets :
        ipos = ncycles * testsize
        testsubs = indexes[ipos:ipos+testsize]
        testset = set(testsubs)  ##  could include repetitions
        while len(testset) < testsize:  ##  rare but possible
            ##  print(len(testsubs),testsubs)
            testset.add(random.choice(rownums))
            ##  print(testset)
        testsubs = list(testset)
        used.update(testset)
        assert len(testsubs) == testsize
        ##  print(testsubs)
        ##  training sample consists of rows absent from testsubs :
        paradat.testsubs = testsubs
        notused = notused - set(testsubs)
        training = []
        for j in range(paradat.rows):
            if not (j in testsubs):
                training.append(j)
        if ncycles < 0:  ##  sanity check
            pass
        paradat.training = training
        ##  print("tr",len(training))
        assert len(training) == paradat.rows - testsize
        ##  generate ruleset :
        rulepop = initpop(paradat)
        ##  bestrule = pickbest(paradat,rulepop)
        bestrule = evotries(paradat,rulepop)
        if ncycles == 0 or bestrule.quality > topscore:
            bout = rulelib.ruleshow(paradat,bestrule,paradat.targcuts)
            for b in bout:
                print(b)
            topscore = bestrule.quality
        ##  for j,item in enumerate(rulepop):
            ##  print(j) ; showobj(item)
        if not hasattr(bestrule,"priorvec"):
            bestrule.priorvec = paradat.priorvec
        elif len(bestrule.priorvec) < 2:
            bestrule.priorvec = paradat.priorvec
        ##  priorvec not added by makerule (carries xbar,sd info).
        if not hasattr(bestrule,"catlist"):
            bestrule.catlist = paradat.targcuts
        elif len(bestrule.priorvec) < 2:
            bestrule.catlist = paradat.targcuts
        ##  likewise with catlist (carries cutpoint into).
        toprules.append(rulelib.copyrule(bestrule))
        ##  treetidy only 2b used after subsampling cycles.
        ##  test ruleset on test cases :
        testrex = regtest(paradat,bestrule,testsubs,"testmode")
        ##  record results :
        results.extend(testrex)
        ncycles += 1
        print(ncycles,cycling,"done.")
        cycling = "cycles"
        if ncycles > testnum:
            break
    if len(notused) < 22:
        print("unused test cases :",sorted(notused))
    else:
        print("unused test cases =",len(notused))
    print("cases used at least once =",len(set(used)))
    nres = len(results)
    print(ncycles,"cycles,",nres,"subsampling instances.")
    paradat.nres = nres
    ##  summarize results :
    dustbins = hh = hits = 0
    for tr in sorted(results,reverse=True):
        print(tr,file=paradat.dumpfile)
        if tr[-1] == Goodsign:
            hits += 1
        if tr[1] < 10:  ##  small cellsize
            dustbins += 1
        else:
            if tr[-1] == Goodsign:
                hh += 1
    hitrate = 100*hits/nres
    print("overall hitrate % =",round(hitrate,4))
    if paradat.rulemode == "tabular":
        dustrate = 100*dustbins/nres
        print("percentage small-sample decisions =",round(dustrate,4))
        hr = 100*hh / max(1,nres-dustbins)
        print("hitrate % omitting small-samples =",round(hr,4))

    ##  pick rule that does best on full dataset :
    print(len(rownums))
    paradat.training = rownums  ##  evaluate on full dataset
    if paradat.rulemode != "tabular":  ##  choose appropriate fitness function
        ff = rulelib.regscore
    else:
        ff = rulelib.qualitab  ##  slow version at finale
    toprule = topbest(paradat,toprules,ff)
    ##  better to do evotries on full dataset ?
    tops = rulelib.ruleshow(paradat,toprule,paradat.targcuts)
    print("....")
    for topline in tops:
        print(topline)
    if paradat.rulemode != "tabular":
        ##  temp check :
        for j in rownums:
            ##  break  ##  currently disabled
            rowvals = paradat.datamat[j]
            v = treelib.treeval(toprule.testlist[0],paradat,rowvals)
            print(j,v,paradat.targvals[j],file=paradat.dumpfile)
    paradat.dumpfile.close()
    
    ##  dump rule for external usage :
    rulelib.ruledump(paradat,tops,paradat.rulefile)
    rf = os.path.abspath(paradat.rulefile.name)
    paradat.rulefile.close()
    
    ##  o/p of results :
    nout = showregs(paradat,results,toprule,"subsampling")
    print("cases processed =",nout)
    print(NL,file=paradat.listfile)  ##  insert blanks
    parashow(paradat,paradat.listfile)
    fn = os.path.abspath(paradat.listfile.name)
    print("output listing on :",fn)
    paradat.listfile.close()
    print("ruleset dumped on :",rf)
    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    pybyebye()
    
    return  0




##  TARGETS :
def targets (paradat,enstring=False):
    "sets up .targvals from .targval expression & updates .cols2use."

    tx = paradat.targval.strip()
    paradat.goodlist = sorted(paradat.goodcols)
    allcols = set(range(paradat.cols))
    paradat.strcols = sorted(allcols - paradat.goodcols)
    paradat.cols2use = allcols
    paradat.targvals = []
    if tx in paradat.taglist:  ##  simple variable
        paradat.targcol = paradat.taglist.index(tx)
        for rowvals in paradat.datamat:
            tv = rowvals[paradat.targcol]
            if enstring:
                tv = str(tv)
            paradat.targvals.append(tv)
        paradat.cols2use.discard(paradat.targcol)
    else:  ##  presume expression
        print("Evaluating target expression : ",tx)
        ##  convert to internal tree format :
        toks = treelib.treetox(tx)
        print(toks)
        pole = treelib.tox2pole(toks)
        ##  print(pole)
        targtree = treelib.poletree(paradat,pole)
        for rowvals in paradat.datamat:
            tv = treelib.treeval(targtree,paradat,rowvals)
            if enstring:
                tv = str(tv)  ##  hold as string
            paradat.targvals.append(tv)
        ##  remove variables in target expression :
        nodelist = treelib.getnodes(targtree,paradat)
        for n in nodelist:
            ##showobj(n)
            if n.type == "feature":
                paradat.cols2use.discard(n.item)
            elif n.type == "stringy":
                paradat.cols2use.discard(n.item[0])  ##  first part of pair is colnum

    return


##  CYCLIST :
def cyclist (rows,nicesubs=144,nicereps=8):
    "tries to compute suitable number of subsampling cycles."

    testsize = int(round(rows**0.5))
    print(':',rows,testsize)

    bestreps = int(round(10.0 - math.log10(rows)))  ##  !!
    bestreps = max(2,bestreps)  ##  never less than 2
    
    return  bestreps,testsize


##  REGMARK :
def regmark (paradat,yhat,yval,fivenums):
    "uses Tukey five-nums to decide 'success' of prediction."

    k = khat = 0
    for y in fivenums[:2]:
        if yhat >= y:  k += 1
        if yval >= y:  khat += 1
    for y in fivenums[2:]:
        if yhat > y:  k += 1
        if yval > y:  khat += 1

    if k == khat:
        return  Goodsign

    return  Poorsign  ##  ybar should be in fivenums if len(fivenums) < 5.


##  FILLTEST :
def filltest (paradat):
    "initial test of table filling in tabular regression mode."

    rr = rulelib.makerule(paradat)
    ts = "((bombing + headroom) > 4.4)"
    rr.testlist[0] = treelib.str2tree(paradat,ts)
    ##  rulelib.regfill(paradat,rr)
    rq = rulelib.qualitab(paradat,rr)
    rr.quality = rq
    outlines = rulelib.ruleshow(paradat,rr,paradat.targcuts)
    for rout in outlines:
        print(rout)

    return


##  STABPREP :
def stabprep (paradat):
    "computes sigtable size & saves info in paradat."

    ##  compute desired number of subrules :
    if paradat.rulemode != "tabular":  ##  assume standard
        paradat.subrules = 1
        paradat.stabrows = 1
    ##  elif paradat.rulemode == "demonic":  ##  homage to Pandemonium
        ##  demonic regression is conceivable !
        ##  as is sequential !!
    else:  ##  default is standard mode
        ##  sr = 1
        ##  while 2**sr <= len(paradat.targcuts):
            ##  sr += 1
        sr = 1
        if paradat.cats > 9:
            sr = 2  ##  plenty of values
        while sr < 9:  ##  huge tables too slow
            stabrows = 2 ** sr
            cellmean = paradat.rows / stabrows
            ##  print(sr,stabrows,cellmean)
            if cellmean < 32:
                break
            sr += 1
        paradat.subrules = sr
        paradat.stabrows = 2**sr

    print("subrules =",paradat.subrules)  ##  same as len(testlist)
    print("stabrows =",paradat.stabrows)
    
    return


##  TREETEST :
def treetest (paradat):
    "dumps treelib results for checking."

    funx = [treelib.makebool,treelib.makecomp,treelib.makemath,treelib.maketree]
    print(paradat.taglist,file=paradat.dumpfile)
    for r,rowvals in enumerate(paradat.datamat):
        print(rowvals,file=paradat.dumpfile)
        funk = random.choice(funx)
        tx = funk(paradat,random.randint(4,7))
        v = treelib.treeval(tx,paradat,rowvals)
        st = treelib.tree2str(tx,paradat)
        print(st,file=paradat.dumpfile)
        print(v,file=paradat.dumpfile)

    return


##  STRCHECK :
def strcheck (paradat):
    "dumps string info for checking."


    print(paradat.taglist,file=paradat.dumpfile)
    for r,rowvals in enumerate(paradat.datamat):
        print(rowvals,file=paradat.dumpfile)
        for sc in paradat.strcols:
            print(sc,paradat.taglist[sc],file=paradat.dumpfile)
            for s in paradat.strdict[sc]:
                v = s in rowvals[sc]
                print(s,v,file=paradat.dumpfile)

    return


##  PICKCUTS :
def pickcuts (paradat):
    "finds probably useful cutpoints for numeric vars."
    
    cutvals = {}
    for vn,varname in enumerate(paradat.taglist):
        if vn in paradat.goodlist:  ##  should be numeric
            ##  extract column vector :
            xvec = [paradat.datamat[j][vn] for j in range(paradat.rows)]
            cuts = getcuts(xvec)
            cutvals[varname] = cuts
            ##  print('#',vn,varname)  ##,file=paradat.dumpfile)
            ##  lout = ''
            ##  for cutpoint in cuts:
                ##  lout = lout + str(cutpoint) + Htab
            ##  print(lout[:-1])  ##,file=paradat.dumpfile)
    paradat.cutvals = cutvals
    ##  also compute targstat :
    xvec = paradat.targvals
    paradat.targstat = bastatz(xvec)
    paradat.targstat.madm = madm(xvec)
    paradat.targstat.midvalue = midvalue(xvec)

    return


##  PICKSTRS :
def pickstrs (paradat):
    "picks constants for string variables."

    n = paradat.rows
    nstrings = round(math.sqrt(n) + math.log1p(n))
    print("desired number of string constants :",nstrings)
    paradat.nstrings = nstrings
    ##  desired number of string constants for each string variable.
    ##  find string fields :
    strcols = []
    for j,t in enumerate(paradat.taglist):
        nn = paradat.tagdict[t]
        if nn < n-1:  ##  nonnumeric
            strcols.append(j)
            ##  print(t)
    paradat.strcols = strcols
    paradat.strdict = {}
    for sc in strcols:
        strvals = seekstrs(paradat,sc)
        print(sc,paradat.taglist[sc],file=paradat.dumpfile)
        print(len(strvals),strvals,file=paradat.dumpfile)
        paradat.strdict[sc] = []
        for sv in strvals:
            s = sv[-1]  ##  last item is substring itself
            paradat.strdict[sc].append(s)  ##  include in vocabulary for sc
            
    return


##  SEEKSTRS :
def seekstrs (paradat,gridcol):
    "returns suitable list of string constants for a grid column."

    ##  first try2 use whole strings :
    frex = getvocs(paradat,gridcol)  ##  already sorted in desc order
    if len(frex) <= paradat.nstrings:  ##  just use them all
        return  frex
    ##  what if 1 string covers all ?? (return null)
    freqtot = 0 ; tops = 0
    enuf = paradat.rows / Gold
    ns = min(len(frex),paradat.nstrings+1)
    for j in range(ns):
        f,s = frex[j]
        freqtot += f
        if freqtot >= enuf:  ##  enough
            tops = j + 1
            print(j,s,freqtot)
            break
    ##  print(j,tops)
    if tops > 0:
        print("success.")
        return  frex[:tops]
        ##  remove hapaxes & uniforms ?
    ##  next try2 find substrings :
    subs = seeksubs(paradat,gridcol)
    if len(subs) < 2:  ##  useless
        return  []
    ##  here try2 pick a suitable number :
    ##  print(len(subs),subs)
    bestsubs = nicesubs(paradat,gridcol,subs)

    return  bestsubs


##  NICESUBS :
def nicesubs (paradat,gridcol,subs):
    "selects probably useful substrings from subs for strings in gridcol."

    ns = len(subs)  ##  already known to be more than 1.
    if ns <= paradat.nstrings:
        return  subs  ##  suitable number
    ##  greedy loop to seek full coverage :
    covered = set()
    newsubs = []
    ncycles = min(ns,paradat.nstrings*2)
    for j in range(ncycles):  ## don't loop for ever
        c = len(covered)  ##  covered already
        uc = paradat.rows - c  ##  still 2b covered
        sortlist = []
        ##  loop thru all substrings :
        for sp,item in enumerate(subs):
            s = item[-1]  ##  last is string itself
            hits = 0
            for r in range(paradat.rows):
                if r in covered:
                    continue
                if s in paradat.datamat[r][gridcol]:
                    hits += 1
            frac = hits / uc
            sortlist.append((frac*(1.0-frac),len(s),hits,s))  ##  nearest to .5 is best
        sortlist.sort(reverse=True)
        str1 = sortlist[0]
        ##  print(j)
        for jj in range(min(0,len(sortlist))):
            print(sortlist[jj])
        if str1[0] < 0.015625:
            break  ##  no use
        newsubs.append((str1[-2],str1[-1]))
        if len(newsubs) >= paradat.nstrings:
            break
        ##  update coverage :
        s = newsubs[-1][-1]
        for r in range(paradat.rows):
            if s in paradat.datamat[r][gridcol]:
                covered.add(r)
        if len(covered) > paradat.rows - 2:
            break  ##  full coverage already

    return  newsubs


##  SEEKSUBS :
def seeksubs (paradat,gridcol):
    "seeks random substrings as constants for a grid column."

    vset = set()
    rows = paradat.rows
    ##  probe at random :
    for p in range(1024):
        if len(vset) >= 987:
            break
        r = random.randint(0,rows-1)
        s = paradat.datamat[r][gridcol]
        rx = random.randint(1,4)
        if rx == 1:  ##  treat s as whole
            vset.add(s)
        elif rx == 4:  ##  try2 find words in s
            sl = len(s)
            sx = [' '] * sl
            for jj in range(sl):
                if s[jj].isalnum():
                    sx[jj] = s[jj]
            sx = ''.join(sx)
            bits = sx.split()
            b = random.choice(bits)
            vset.add(b)
        else:  ##  extract substring
            sl = len(s)  ##  what if sl is zero ?!
            hi = min(9,sl)
            i = random.randint(0,sl-1)  ##  start
            ix = random.randint(1,hi)  ##  size
            slab = s[i:i+ix]
            if slab:
                vset.add(slab)
    ##  next tally them :
    frex = []
    for slab in vset:
        if not slab or (Backtick in slab) or (Quotes2 in slab):
            continue  ##  unsuitable
        ns = 0
        for r in range(rows):
            sv = paradat.datamat[r][gridcol]
            if slab in sv:
                ns += 1
        if ns > 1 and ns < rows:  ##  omit hapaxes & ever-presents
            frex.append((ns,slab))
    frex.sort(reverse=True)

    return  frex


##  GETVOCS :
def getvocs (paradat,gridcol):
    "returns a frequency ordered vocabulary list from grid column."

    vtab = {}
    n = paradat.rows
    for j in range(n):
        s = paradat.datamat[j][gridcol]
        if s in vtab:
            vtab[s] += 1
        else:
            vtab[s] = 1
    freqlist = []
    for s in vtab:
        if s and not ((Backtick in s) or (Quotes2 in s)):
            freqlist.append((vtab[s],s))
            ##  don't want null strings; can't handle grave or double quotes.

    return  sorted(freqlist,reverse=True)


##  SLIMMER :
def slimmer (paradat,rule):
    "makes copy of rule with tidied tests."

    sr = rulelib.copyrule(rule)
    for j in range(len(rule.testlist)):
        sr.testlist[j] = tidylib.treetidy(paradat,sr.testlist[j])

    return  sr


##  TOPBEST :
def topbest (paradat,toprules,ff):
    "returns rule that does best on full dataset."

    fittest = toprules[0]
    topscore = [-Huge]
    ##  revise fallout tables, hence fitnesses :
    for rule in toprules:
        slimrule = slimmer(paradat,rule)
        rulelib.regfill(paradat,rule)  ##  assume .training already updated
        rulelib.regfill(paradat,slimrule)
        rule.quality = q = ff(paradat,rule)  ##  fitness function passed in
        slimrule.quality = qq = ff(paradat,slimrule)
        ##  if q != qq:
            ##  print("*!!*",q,qq)
        print('*',file=paradat.dumpfile)
        rout = rulelib.ruleshow(paradat,rule,paradat.targcuts)
        for outline in rout:
            print(outline,file=paradat.dumpfile)
        rout = rulelib.ruleshow(paradat,slimrule,paradat.targcuts)
        for outline in rout:
            print(outline,file=paradat.dumpfile)
        if q > topscore:
            topscore = q
            fittest = rule  ##  safer with rulelib.copyrule() ??
        if qq >= topscore:  ##  prefer when brevity < 1
            topscore = qq
            fittest = slimrule  ##  safer with copyrule() ??
    
    return  fittest


##  SHOWREGS :
def showregs (paradat,results,decrule,h):
    "writes eval info about results (assumed already sorted) onto file."
    ##  version for regression rule mode.
    
    ##  first some parameter info :
    for po in Outpars:
        pv = getattr(paradat,po,'')
        if pv:
            print(po.ljust(10),pv,file=paradat.listfile)
    print("\n===="+h+" trial :\n",file=paradat.listfile)
    
    ##  next headings for subsampling results :
    h1 = "rank  safeness case  #id#"
    h2 = "pred:true      cellsize      abdsiff     diffsqrd"
    vo = ''
    if paradat.idfield and (paradat.idfield in paradat.taglist):
        pi = paradat.taglist.index(paradat.idfield)
        vo = paradat.idfield[:Longname]
        h1 = h1.replace("#id#",vo)
    n = len(h1)
    h = h1.ljust(51) + h2
    print(h,file=paradat.listfile)

    ##  testmode results :
    marklist = []
    results.sort(reverse=True)
    ##  sort by predval (temporary??) :
    resx = [[r[-2][0]]+r for r in results]  ##  should be predval
    ##  print(resx[0])
    resx.sort(reverse=True)
    results = [rx[1:] for rx in resx]
    ##  print(results[0])
    
    rightvec = [] ; wrongvec = []  ##  right/wrong vecs for Hedges's g
    predvec = [] ; truevec = []  ##  for correlations
    svec = []  ##  'safeness' vector
    tot1 = tot2 = 0.0
    nres = len(results)
    for j,r in enumerate(results):
        strength = r[0] ; svec.append(strength)
        cellsize = r[1]
        ix = r[2]
        cellvals = r[3]
        ##  minimal checking !
        predvals = r[4]
        ##  stabpos is at [5].
        predval,trueval = r[-2]
        predvec.append(predval)
        truevec.append(trueval)
        mark = r[-1]
        if mark == Goodsign:  ##  mark already computed
            rightvec.append(strength)
        else:
            ##  mark = Poorsign
            wrongvec.append(strength)
        d = abs(predval-trueval)
        tot1 += d ; tot2 += d*d
        predval = round(predval,4)  ## save space (better format?)  
        marklist.append(mark)
        rowvals = paradat.datamat[ix]
        if vo:
            h = str(rowvals[pi])  ##  case label
        else:
            h = str(ix)  ##  just use serial position
        h = h[:Longlab].ljust(Longlab)  ##  case label
        ##  print(j,strength,h,predval,mark,trueval,cellsize,predvals,
              ##  file=paradat.listfile)
        if j < 0:  ##  checking
            print(r)
            print([j+1,strength,ix,h,predval,mark,trueval,cellsize,[d,d*d]])
        showline(paradat,j+1,strength,ix,h,predval,mark,trueval,cellsize,
                 [d,d*d])

    marx = ''.join(marklist)
    print(marx,file=paradat.listfile)

    sr = len(rightvec) / max(1,nres)
    print("\n'success' percentage =",round(sr*100.0,2),
          file=paradat.listfile)
    cs = corsums(predvec,truevec)
    c = corrval(cs)
    print("pearson correlation between predicted & true vals =",
          round(c,4),file=paradat.listfile)
    cs = corsums(ranx(predvec),ranx(truevec))
    c = corrval(cs)
    print("spearman rank-correlation between predicted & true vals =",
          round(c,4),file=paradat.listfile)
    print("\nmean abs.error =",round(tot1/nres,4),file=paradat.listfile)
    print("mean error ^ 2 =",round(tot2/nres,4),file=paradat.listfile)

    ##  also differential of right & wrong strengths :
    ##  Hedges's g misleading, removed.
    s = "safeness"
    ##  corr with abserror used instead :
    abserror = [abs(predvec[j]-truevec[j]) for j in range(len(results))]
    cs = corsums(abserror,svec) ; c = corrval(cs)
    print("correlation between",s,"& abs.error (negative better) = ",
          end='',file=paradat.listfile)
    print(round(c,4),file=paradat.listfile)
    ##  print(sum(rightvec)/len(rightvec))
    ##  print(sum(wrongvec)/len(wrongvec))

    ##  also show (intelligible?) version of decision rule :
    rout = rulelib.ruleshow(paradat,decrule,paradat.targcuts)
    print("\nResultant rule from all training cases :",file=paradat.listfile)
    for outline in rout:
        print(outline,file=paradat.listfile)

    return  len(results)


##  SHOWLINE :
def showline (paradat,rank,strength,ix,caselab,predval,mark,trueval,cellsize,
              predvals):
    "displays single line on paradat.listfile."

    fm = "{0:4d}  {1:8.2f} {2:4d}  {3:20s} {4:>12s} ~ {5:12s} {6:4d}"
    fm = fm.replace('~',mark,1)
    pv = str(predval)[:12] ; tv = str(trueval)[:12]
    out1 = fm.format(rank,strength,ix,caselab,pv,tv,cellsize)
    out2 = ''
    w = 12
    for pv in predvals:
        out2 = out2 + " {0:{1}.2f}".format(pv,w)
    print(out1+out2,file=paradat.listfile)


##  OUTSHOW :
def outshow (paradat,cmat,seen,fo):
    "shows additional stats."

    ##  display indicators :
    print("\ncases =",rc,file=fo)
    print("cases with unseen category labels =",rc-seen,file=fo)
    if seen > 0:
        if seen == 1:
            word = "case"
        else:
            word = "cases"
        if seen < rc:
            print("cases with known category labels =",seen,file=fo)
            print("[results below only apply to",seen,word+']',file=fo)
        print("hits = ",hits,file=fo)
        percent = 100 * hits / seen
        print("percent hits =",round(percent,2),file=fo)
        
    return


##  RANDSUBS :
def randsubs (ns,subsize,reps):  ##  plus global Mintries
    "returns randomized list of subscripts 2use in subsampling."

    assert subsize <= ns  ##  subsize no larger than number of cases
    ilength = reps * subsize  ##  number of indices that will be needed
    allsubs = list(range(ns))
    random.shuffle(allsubs)
    indices = allsubs[:ilength]  ##  separate copy
    while len(indices) < ilength:
        ix = random.sample(allsubs,subsize)
        indices.extend(ix)

    return  indices  ##  [:ilength] ??  (longer doesn't matter?)


##  REGTEST :
def regtest (paradat,thisrule,testsubs,testmode,tiny=1e-8):
    "applies regression-rule to (potentially) unseen test cases."

    reslist = []
    dx = paradat.smalldif  ##  removable ?
    s = 0
    for t in thisrule.testlist:
        s += treelib.size(t)
    s = s / len(thisrule.testlist)  ##  mean size
    ##  tbar = paradat.targstat.mean  ##  overall mean
    tbar = thisrule.priorvec[1]
    sd = thisrule.priorvec[2]
    for j in testsubs:
        rowvals = paradat.datamat[j]
        stabpos,cellvals = rulelib.evalregs(paradat,thisrule,rowvals)
        if paradat.rulemode != "tabular":
            predval = cellvals[0]  ##  just 1 value in direct mode
        else:  ##  assume tabular
            predval = cellvals[1]
            cellsize = cellvals[0]
            if testmode == "suppress":
                predval = paradat.targnorm + predval
        ##  trueval = rowvals[paradat.targcol]
        trueval = paradat.targvals[j]
        d = abs(predval-tbar)
        d2 = (predval-trueval) ** 2
        if paradat.rulemode != "tabular":
            strength = sd / (sd+d)
            cellsize = paradat.rows
        else:  ##  tabular
            if testmode == "suppress":
                strength = sd/(sd+d)
            else:
                strength = sd / (sd + cellvals[2])
        ##  mark = Poorsign
        ##  print(predval,type(predval))
        ##  print(type(paradat.targcuts[0]))
        ##  mark if 'close' to true value :
        mark = regmark(paradat,predval,trueval,thisrule.catlist)
        ##  print(trueval-predval,mark)
        reslist.append([strength,cellsize,j,cellvals,[predval],stabpos,
                        (predval,trueval),mark])

    return  reslist


##  GETCUTS :
def getcuts (xvec):
    "returns list of suitable cutpoints for numeric column vector."

    n = len(xvec)
    if n < 2:
        return  []
    ##  compute desired number of cutpoints :
    ##  nthresh = math.log1p(n)  ##  sqrt rather too many ?
    nthresh = math.sqrt(n)
    xbar = sum(xvec) / n  ##  mean will go in as well
    xvec.sort()  ##  must be ordered
    p = 0
    vals = []
    incr = n / nthresh
    ##  print(nthresh,incr)
    while p < n:
        ip = int(p)
        vals.append(xvec[ip])
        p += incr
    ##  also ensure presence of Q1,Q2,Q3,Q4 & mean :
    vals.append(xvec[int(n/4)])
    vals.append(xvec[int(n/2)])
    vals.append(xvec[int(n*0.75)])
    vals.append(xvec[-1])
    vals.append(round(xbar,7))
    vals = sorted(set(vals))
    ##  print(len(set(xvec)),len(vals))  ##  how many distinct data values

    return  vals


##  GETFILE :
def getfile (paradat,fi):
    "obtains data grid from file."

    paradat.datamat,paradat.taglist,paradat.tagdict = matread(fi,paradat)
    paradat.rows = len(paradat.datamat)
    print("data rows =",paradat.rows)
    paradat.cols = len(paradat.taglist)
    print("data cols =",paradat.cols)
    print("column names :")
    print(paradat.taglist)
    if paradat.rows < Mincases:  ##  not enough data
        print("Not enough data cases !")
        print("Program halts.")
        return  -1
    goodcols = cols2use(paradat)
    paradat.goodcols = goodcols
    paradat.goodlist = sorted(paradat.goodcols)
    if len(goodcols) < 1:
        print("No numeric variables found in ",paradat.datfile,'!')
        print("Program halts.")
        return  -2

    return  0  ##  success, paradat altered




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
